/*
	C Implementation of the Nearest Neighbour Scaling Algorithm
	for the MEGA Adventure Vision Demo

	by sy2002 on Mar 11 2013
	improved on Mar 16 2013
	
	This code is the basis for the assembler implementation and it is
	optimized to use as less multiplications and divisions as possible
	and as less 16bit stuff as possible due to the 8048 CPU.
	
	Since we use loopup-tables for the width and height ratios, the
	algorithm only needs two 16bit additions and one 8bit MUL.
		
	HOW-TO COMPILE: Use the GNU Compiler Collection:
	g++ -o algorithm algorithm.c
	
	Inspiration for the algorithm taken from:
	http://tech-algorithm.com/articles/nearest-neighbor-image-scaling/
*/

#include <stdio.h>
#include <stdlib.h>

const int w1 = 25;				//source image width
const int h1 = 5;				//source image height

//#define OUTPUT_LOOKUP_TABLES	//define to print the LUTs needed for the ASM
//#define OUTPUT_ANALYTICS		//define to print intel about multipliers and ratios

//source image: MEGA logo: 25 x 5 bytes, 1 = pixel set, 0 = pixel cleared
int pixels[w1 * h1] = 
{
	0, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 
	1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 
	1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 
	1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 
	0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0 
};

int main(int argc, const char* argv[])
{
	printf("Mega Zoomer Algorithm Tester by sy2002 on Mar 16 2013\n");

	//target width/height (zoomed width/height), initialized with the source values
	int w2 = w1;
	int h2 = h1;

	if (argc < 2)
		printf("Specify new width and height as command line parameters.\n");
	else
	{
		w2 = atoi(argv[1]);
		h2 = atoi(argv[2]);
	}
	
	printf("\n");
	
	//zoomed image pixel array
	int* zoom = new int[w2 * h2];

	//the desired ratios should be put to a lookup-table for the AV ASM
	//we do not want 16bit divisions on AV :-)
	int x_ratio = (int)((w1 << 8) / w2) + 1;
	int y_ratio = (int)((h1 << 8) / h2) + 1;
	
	
#ifdef OUTPUT_LOOKUP_TABLES
	//Generate AV lookup-table for ratios
	//use the given w1, h1 sizes (see above) and walk through
	//the below-mentioned start/end values
	const int x_start	= 25;
	const int x_end 	= 50;
	const int y_start	= 5;
	const int y_end		= 10;

	printf("x_ratios from %i to %i for given width %i:\n", x_start, x_end, w1);
	for (int i = x_start; i <= x_end; i++)
	{
		int n = (int)((w1 << 8) / i) + 1;
		printf(".db $%02X, $%02X\t; x = %i, x_ratio = %i\n", i, n, i, n);
	}
	
	printf("y_ratios from %i to %i for given height %i:\n", y_start, y_end, h1);
	for (int i = y_start; i <= y_end; i++)
	{
		int n = (int)((h1 << 8) / i) + 1;
		printf(".db $%02X, $%02X\t; y = %i, y_ratio = %i\n", i, n, i, n);
	}
#endif
	
	// *****************************************************************************************
	// 	THE ALGORITHM FOR ADVENTURE VISION: START
	// *****************************************************************************************
	
	int x2, y2;
	int* t1 = zoom;	
	int y_ratio_mult = 0;

	for (int i = 0; i < h2; i++)
	{	
		//y_ratio_mult is 16bit, but >> 8 means basically just: take the hi-byte of y2
		y2 = (y_ratio_mult >> 8);
		
		//will always be < h2, therefore 8-bit mul is sufficient
		//due to the screen size constraints of our 50x40 mode
		int* p = pixels + (y2 * w1);

#ifdef OUTPUT_ANALYTICS
		//output multiples for debugging and analysis purposes
		printf("y_ratio_mult = %i\ty2 = %i\t\ty2 * w1 = %i\n", y_ratio_mult, y2, (y2 * w1));
#endif

		int rat = 0;
		int* t2 = t1;
		
		for (int j = 0; j < w2; j++)
		{			
			x2 = (rat >> 8);
			
			//in asm this will be putpixel with (r0|r1) = (j|i), i.e. no t1, t2 needed
			*t2 = *(p + x2);
			t2++;
			
			//this is a 16bit add
			rat += x_ratio;
		}
		
		//y_ratio_mult is 16bit, i.e. a 16bit add is needed
		y_ratio_mult += y_ratio;
		
		//the t1, t2 arithmetic can be skipped in asm; replaced by putpixel
		t1 += w2;
	}
	
	// *****************************************************************************************
	// 	THE ALGORITHM FOR ADVENTURE VISION: END
	// *****************************************************************************************

	//print the scaled logo to the terminal
	for (int y = 0; y < h2; y++)
	{
		for (int x = 0; x < w2; x++)
			if (zoom[(y * w2) + x] == 1)
				printf("o");
			else
				printf(" ");
		
		printf("\n");
	}
	
	printf("\n");
	delete[] zoom;
}
